__all__ = ("Loader",)

from pathlib import PurePath

from .file_ import Loader as FileLoader, LocaleFile

try:
    from orjson import loads
except ModuleNotFoundError:
    from json import loads


class Loader(FileLoader):
    """Loader implemented to load json files.
    By default, json is read using the standard module json.
    But if you have ``orjson`` or ``i21y[fast-json]`` installed, json reading is done with the library called orjson.
    The arguments of constructor are same as :class:`.file_.Loader`."""

    EXTENSIONS = (".json",)

    def load(self, path: PurePath) -> LocaleFile:
        lf = super().load(path)
        with open(lf.path, "rb") as f:
            lf.data = loads(f.read())
        return lf



from src.i21y import Translator
from src.i21y.loaders.yaml import Loader


def do_test_translator(t: Translator[Loader]) -> None:
    assert t("main.test") == "Test"
    assert t("main.test", locale="ja") == "テスト"

    assert t("main.a.b.c") == "See"
    assert t("main.a.b.c", locale="ja") == "しー"

    assert t("main.a.b.d.e") == "Yee"
    assert t("main.a.b.d.e", locale="ja") == "いー"

    assert t("main.a.b.f") == "Eff:ect"
    assert t("main.a.b.f", locale="ja") == "エフ：ェクト"


do_test_translator(Translator(Loader("tests/locale")))



from __future__ import annotations

__all__ = ("Translator",)

from typing import TYPE_CHECKING, TypeVar, Generic, Any, cast

if TYPE_CHECKING:
    from .abc import Loader
    from .utils import locale_str


LoaderT = TypeVar("LoaderT", bound="Loader")


class Translator(Generic[LoaderT]):
    """This class is for translation.

    Args:
        loader: It is for searching translations.
        default_locale: It will be used for default locale. If you don't specify it, it will be set `en`.
    """

    def __init__(self, loader: LoaderT | Loader, default_locale: str = "en") -> None:
        self.loader: LoaderT = cast(LoaderT, loader)
        self.default_locale = default_locale

    def format_(self, text: str, *args: Any, **kwargs: Any) -> str:
        """Format text.

        Args:
            text: Target text.
            *args: The arguments to be passed to ``text.format``.
            **kwargs: The keyword arguments to be passed to ``text.format``."""
        return (
            text.format(*args, **kwargs)
            if kwargs.pop("__i21y_special_kwarg_do_format", True)
            else text
        )

    def translate(
        self,
        key: str | locale_str,
        *args: Any,
        locale: str | None = None,
        do_format: bool = True,
        **kwargs: Any,
    ) -> str:
        """Do translation.
        The translation will be formated by :meth:`Translator.format_`.

        Args:
            key: It will be used for searching text.
            *args: The arguments to be passed to :meth:`.format_`.
            locale: The locale of text.
            do_format: Whether or not to perform formatting.
            **kwargs: The arguments to be passed to :meth:`.format_`.

        Note:
            It can be called by calling instance of :class:`Translator`. Like:

            .. code-block:: python

                t = Translator()
                print(t("responses.not_found"))

        Raises:
            TranslationNotFound: If translation is not found, this function will raise it.
        """
        kwargs["__i21y_special_kwarg_do_format"] = do_format
        return self.format_(
            self.loader.search(
                locale or self.default_locale, key if isinstance(key, str) else str(key)
            ),
            *args,
            **kwargs,
        )

    def __call__(self, key: str | locale_str, *args: Any, **kwargs: Any) -> str:
        return self.translate(key, *args, **kwargs)

    def __str__(self) -> str:
        return f"<Translator loader={self.loader} default_locale={self.default_locale}>"



from __future__ import annotations

__all__ = ("LocaleFile", "Loader")

from typing import NamedTuple, TypeAlias, cast
from collections.abc import Iterable

from collections import defaultdict
from functools import cache

from pathlib import PurePath
from os.path import exists, isfile
from os import walk

from ..abc import Loader as AbcLoader, SearchT
from ..error import TranslationNotFound
from ..utils import Undefined


DataDict: TypeAlias = dict[str, "str | DataDict"]


class LocaleFile:
    """Class for handling translation data file.

    Args:
        parent: The source from which the instance was created.
        locale: The locale of the file.
        path: The path to the file."""

    def __init__(self, parent: Loader, locale: str, path: PurePath) -> None:
        self.parent, self.path, self.data = parent, path, DataDict()
        self.locale, self.key = locale, ".".join(
            self.path.parts[parent.path_size + 1 : -1] + (self.path.stem,)
        )

    def get(self, key: str | Iterable[str]) -> str | None:
        """Search for translations in this file.

        Args:
            key: The key to search for translations."""
        if isinstance(key, str):
            key = key.split(".")

        tentative, result = None, None
        for partial in key:
            if result is not None:
                # なぜ下の`result = ...`ではなくここでわざわざ`return`するかは、既に文字列まで辿り着いたのに深くまでいくということがないようにするためのもの。
                # 例えば、`a.b.c`で文字列があるとき、`a.b.c.d.e`のeまでアクセスがすることができてしまう。
                # ましてや、`a.b.c`の文字列が返却される。これはよろしくないので、わざと一度イテレートしてからreturnをして、呼び出し元にエラーを認識してもらう。
                return

            try:
                if tentative is None:
                    tentative = self.data[partial]
                else:
                    assert not isinstance(tentative, str)
                    tentative = tentative[partial]

                if isinstance(tentative, str):
                    result = tentative
            except KeyError:
                break

        if result is not None:
            return result

    def __str__(self) -> str:
        return f"<LocaleFile path={self.path}>"


class Caches(NamedTuple):
    "Class for storing multiple caches."

    file_: defaultdict[str, dict[str, LocaleFile]]
    """Cache of :class:`LocaleFile`.
    Type is ``collections.defaultdict[locale, dict[key to file, LocaleFile]]``."""
    key_mapping: dict[tuple[str, str], tuple[str, list[str]]]
    """Cache of keys to files.
    Type is this:

    .. code-block::

        dict[
            (locale, key to translation),
            (
                key to file that has translation,
                key without key to the file
            )
        ]"""

    def clear(self) -> None:
        "Clear all cache."
        self.file_.clear()
        self.key_mapping.clear()

    def __str__(self) -> str:
        return f"<Caches text={self.file_} key_mapping={self.key_mapping}>"


class Loader(AbcLoader):
    """Loader class implemented for loading files.
    This is just an environment for loading files, and is not implemented to actually load json files, etc.
    So, if you want to use it in real life, you can either implement it yourself, or use the :class:`.json.Loader` provided by i21y.

    Args:
        path: Path to directory for placeing translations. It is often set to ``locale``.
        preload_cache: Whether to run :meth:`.make_cache` when initializing the instance.
        use_cache_realtime: Whether to search and collect caches when translating.
        do_not_search_file: Whether to prevent file searches from being performed.
            If you do not want IO processing during translation, you can set this to ``True``.
    """

    EXTENSIONS = tuple[str, ...]()
    "A tuple of supported file extensions. This must not be empty."

    def __init__(
        self,
        path: str | PurePath,
        preload_cache: bool = True,
        use_cache_realtime: bool = True,
        do_not_search_file: bool = False,
    ) -> None:
        if isinstance(path, str):
            path = PurePath(path)
        self.path = path
        self.path_size = len(path.parts)
        self.do_not_search_file = do_not_search_file

        self.caches = Caches(defaultdict(dict), {})

        self.use_cache_realtime = use_cache_realtime
        if preload_cache:
            self.make_cache()

        if not self.EXTENSIONS:
            raise NotImplementedError(
                "The file extension is not set. This indicates that this class is not properly implemented."
            )

    def load(self, path: PurePath) -> LocaleFile:
        """Load the file.
        You must implement it to load the data from the file.
        The default implementation returns an instance of :class:`.file_.LocaleFile` with empty data.

        Args:
            path: The path to the file."""
        return LocaleFile(self, path.parts[self.path_size], path)

    def make_cache(self) -> None:
        "Searches for translations and generates a cache."
        for root, _, files in walk(self.path):
            for file_name in files:
                if not file_name.endswith(self.EXTENSIONS):
                    continue

                lf = self.load(PurePath(root).joinpath(file_name))
                self.caches.file_[lf.locale][lf.key] = lf

    @cache
    def get_locale_path(self, locale: str) -> PurePath:
        """Create an instance of :class:`pathlib.PurePath` representing the path to the folder where the translations are located.

        Args:
            locale: The locale of translations."""
        return self.path.joinpath(locale)

    def search_locale_file(
        self, locale: str, key: str
    ) -> tuple[list[str], LocaleFile] | None:
        """Searchs for translations file.

        Args:
            locale: The locale of translations.
            key: The key to translations."""
        # キャッシュを検索する。
        if self.use_cache_realtime:
            # 指定されたキーの文章たあるファイルのキーがキーマッピングのキャッシュにあるかを検索する。
            if (locale, key) in self.caches.key_mapping:
                file_key, end_of_key = self.caches.key_mapping[(locale, key)]
                return end_of_key, self.caches.file_[locale][file_key]

            # 普通に指定されたキーのファイルをキャッシュから検索する。
            for tentative in filter(key.startswith, self.caches.file_[locale].keys()):
                end_of_key = key.replace(tentative, "")

                if end_of_key.startswith("."):
                    end_of_key = end_of_key[1:]

                _, eok = self.caches.key_mapping[(locale, key)] = (
                    tentative,
                    end_of_key.split("."),
                )

                return eok, self.caches.file_[locale][tentative]

        if self.do_not_search_file:
            return

        # 検索を行う。
        path, file_key = self.get_locale_path(locale), ""
        for index, partial in enumerate(keys := key.split(".")):
            file_key += f"{partial}."

            # ディレクトリかどうかをチェックする。
            path = path.joinpath(partial)
            if exists(path):
                continue

            # ファイルかどうかをチェックする。
            for ext in self.EXTENSIONS:
                path = path.with_suffix(ext)

                if exists(path) or isfile(path):
                    # ファイルが存在したのなら、ロードを行う。
                    lf = self.load(path)
                    if self.use_cache_realtime:
                        self.caches.file_[locale][file_key[:-1]] = lf
                    return keys[index + 1 :], lf
            else:
                break

    _TNF_T = "Not found text: %s"

    def search_impl(
        self, locale: str, key: str, default: type[Undefined] | SearchT = Undefined
    ) -> str | SearchT:
        result = self.search_locale_file(locale, key)
        if result is None:
            if default is Undefined:
                raise TranslationNotFound(key, self._TNF_T % key)
            return cast(SearchT, default)
        end_of_key, lf = result
        if (text := lf.get(end_of_key)) is None:
            raise TranslationNotFound(key, self._TNF_T % key)
        return text



from __future__ import annotations

__all__ = ("Undefined", "locale_str")

from typing import TypeAlias, TypeVar, Any, overload
from collections.abc import Iterable


class Undefined:
    "This class is used to represent undefined."


LocaleStr: TypeAlias = "str | locale_str"
SelfT = TypeVar("SelfT", bound="locale_str")
AdtnlClsT = TypeVar("AdtnlClsT", bound="locale_str")


class locale_str:
    """This class is designed to allow for the graceful creation of keys to translations.

    **Operations**

    .. describe:: x + y

        It is alias for :meth:`locale_str.join`.
        y can be specified str or Iterable, and y will be passed to :meth:`.join` as arguments.
    .. describe:: x & y

        It is alias for :meth:`locale_str.join_raw`.
        y can be specified str or Iterable, and y will be passed to :meth:`.join_raw` as arguments.
    .. describe:: x == y

        Checks if :attr:`locale_str.key` of x and y are the same.
    .. describe:: x != y

        Checks if :attr:`locale_str.key` of x and y are not the same.

    Args:
        key: The key to ranslation.
        **extras: It is for some data.

    Notes:
        Attribute can be used as the argument of :meth:`.join`."""

    def __init__(self, key: Any, **extras: Any) -> None:
        self.key, self.extras = key, extras

    def _adjust_key(self, key: str) -> str:
        if not self.key.endswith("."):
            key = f".{key}"
        return key

    def join_raw(self, *other: str | locale_str) -> str:
        """Concatenate strings to self and return string of it.

        Args:
            *other: Target strings."""
        new = self.key
        for partial in other:
            if not isinstance(partial, str):
                partial = str(partial)
            new = f"{new}{self._adjust_key(partial)}"
        return new

    @overload
    def join(
        self: ..., *other: str | locale_str, cls: type[AdtnlClsT], **extras: Any
    ) -> AdtnlClsT:
        ...

    @overload
    def join(
        self: SelfT, *other: str | locale_str, cls: None = None, **extras: Any
    ) -> SelfT:
        ...

    def join(
        self: SelfT,
        *other: str | locale_str,
        cls: type[AdtnlClsT] | None = None,
        **extras: Any,
    ) -> SelfT | AdtnlClsT:
        """A version of :meth:`locale_str.join_raw` that returns the instance of the class that self.

        Args:
            *other: Target strings.
            cls: The :class:`.locale_str` used to create the instance.
            **extras: The value to pass to the ``extras`` argument of the constructor of :class:`.locale_str`.
        """
        cls = cls or self.__class__  # type: ignore
        assert cls is not None
        return cls(self.join_raw(*other), **extras)

    def update_extras(self: SelfT, **extras: Any) -> SelfT:
        """Function to gracefully update :attr:`.extras`.

        Args:
            extras: The extras to override."""
        self.extras.update(**extras)
        return self

    def __add__(self: SelfT, other: LocaleStr | Iterable[LocaleStr]) -> SelfT:
        if isinstance(other, str | locale_str):
            other = (other,)
        return self.join(*other)

    def __and__(self, other: str | Iterable[str]) -> str:
        if isinstance(other, str):
            other = (other,)
        return self.join_raw(*other)

    def __getattr__(self: SelfT, another: str) -> SelfT:
        return self.join(another)

    def __eq__(self, another: locale_str) -> bool:
        return self.key == another.key

    def __ne__(self, another: locale_str) -> bool:
        return self.key != another.key

    def __str__(self) -> str:
        return self.key



from src.i21y import locale_str


BASE = locale_str("aiueo")
assert str(BASE) == "aiueo" and BASE.key == "aiueo"

KAGYO = locale_str("kakikukeko")
RIGHT = BASE + KAGYO
LEFT = BASE + str(KAGYO)
assert LEFT == RIGHT
assert str(LEFT) == str(RIGHT)
assert LEFT.test == RIGHT.test

assert BASE & "kakikukeko" == "aiueo.kakikukeko"

assert BASE.join("test", a="b").extras["a"] == "b"
assert BASE.test.update_extras(a="b").extras["a"] == "b"



from src.i21y import Translator
from src.i21y.loaders.json import Loader


def do_test_translator(t: Translator[Loader]) -> None:
    assert t("general.main") == "main"
    assert t("general.main", locale="ja") == "メイン"

    from src.i21y.utils import locale_str

    BASE = locale_str("wow.a")
    assert t(BASE.b) == "Layor B"
    assert t(BASE.b, locale="ja") == "B階層"

    assert t(BASE.c.d, "test_arg") == "Layor D: test_arg"
    assert t(BASE.c.d, "test_arg", locale="ja") == "D階層：test_arg"

    assert t(BASE.c.d, do_format=False) == r"Layor D: {}"


do_test_translator(Translator(Loader("tests/locale", False, False)))
do_test_translator(Translator(Loader("tests/locale", do_not_search_file=True)))
do_test_translator(Translator(Loader("tests/locale")))



__author__ = "Takagi Tasuku"
__version__ = "0.2.0"
__all__ = ("Translator", "I21YError", "TranslationNotFound", "locale_str")

from .translator import Translator
from .error import I21YError, TranslationNotFound
from .utils import locale_str
